function [ z, K, w_ret, mu_ret, d_ret ] = dpspace( x, lambda, s, a, maxK, maxT, eps, quiet )
% Input:
%   x: real n x D, training data
%   lambda, s, a: real scalar, hyperparameters
%   maxK: int scalar, maximum number of components allowed
%   maxT: int scalar, maximum number of iterations allowed
%   eps: real scalar, convergence control
%
% Output:
%   z: int n x 1, cluster assignment
%   K: int scalar, # of components
%   w_ret: real K x D x D, projection matrices. Only the left d_k columns
%   matter
%   mu_ret: real K x D, offset
%   d_ret: int K x 1, subspace dimensions

    % Set up parameters
    n = size(x, 1);
    D = size(x, 2);

    % Initialization
    z = ones(n, 1);
    K = 1;
    U = zeros(maxK, D, D);
    L = zeros(maxK, D);
    mu = zeros(maxK, D);
    d = zeros(maxK, 1);
    
    Q = zeros(n, maxK+1);
    nk = zeros(maxK, 1);
    
    d(1) = 0;
    mu(1, :) = mean(x);
    U(1, :, :) = eye(D);
    nk(1) = n;
    mu1 = mu(1, :);
    Q(:, 1) = sum((x - mu1(ones(n, 1), :)) .^ 2, 2);
    for k = 2:maxK+1
        Q(:, k) = 1e100 * ones(n, 1);
    end
    
    [loss, lossA, lossB, lossC] = compute_loss(x, z, K, U, L, mu, d, lambda, s);
    if (quiet ~= 1)
        fprintf('Init loss = %f, A = %f, B = %f, C = %f\n', loss, lossA, lossB, lossC);
    end
    lastloss = 1e100;
    
    T = 0;
    while (T < maxT && abs(lastloss - loss) / lastloss > eps)
        
        T = T + 1;
        lastloss = loss;
        
        % update z: only try to create new clusters
        [minQ, z] = min(Q, [], 2); % n x 1
        while (K < maxK && max(minQ) > lambda)
            ind = (minQ > lambda * ones(n, 1)); % n x 1
            [max_value, idx] = max(ind);
            K = K + 1;
            U(K, :, :) = eye(D);
            muk = x(idx, :);
            mu(K, :) = muk;
            d(K) = 0;
            Q(:, K) = sum((x - muk(ones(n, 1), :)) .^ 2, 2);
            [minQ, z] = min(Q, [], 2);
        end
        
     %   [loss, lossA, lossB, lossC] = compute_loss(x, z, K, U, L, mu, d, lambda, s);
     %   fprintf('Iter %d, loss = %f, lossA = %f, lossB = %f, lossC = %f, K = %d\n', T, loss, lossA, lossB, lossC, K);
        
        clean_up;
       
        % update W_k and mu_k
        for k = 1:K
            if (nk(k) == 0)
                continue;
            end

            ind = (z == k * ones(n, 1)); % n x 1
            xind = ind(:, ones(1, D)) .* x; % n x D
            sortQ = sort(Q, 2); % n x maxK+1
            
            % update mu	
            muk = 1.0/nk(k) * sum(xind); % 1 x D
            mu(k, :) = muk;
            
            % update U
            xind_shift = ind(:, ones(1, D)) .* (x - muk(ones(n, 1), :)); % n x D
            S = 1.0/nk(k) * xind_shift' * xind_shift; % D x D
            [Uk, G, V] = svd(S); 
            U(k, :, :) = Uk;
            
            % update dk
            U_xind_shift_2 = (xind_shift * Uk) .^ 2; % n x D
            costs = zeros(1, D);
            for j = 0:D-1
                cost1 = sum(U_xind_shift_2(:, j+1:D), 2); % n x 1
                cost2 = sortQ(:, 2); % n x 1
                %cost2 = 1e100 * ones(n, 1);
                costs(j+1) = s * j + sum(ind .* min(cost1, cost2));
            end
            [min_value, d(k)] = min(costs);
            d(k) = d(k) - 1;
            
            % update L
            for j = 1:d(k)
                L(k, j) = G(j, j) / (1 + 2*(a-1)/D);
            end
            
            % update cost
            U_shift_2 = ((x - muk(ones(n, 1), :)) * Uk) .^ 2;
            Q(:, k) = sum(U_shift_2(:, d(k)+1:D), 2);
            
            % update z (without the possibility of creating a new cluster
          %  [minQ, z] = min(Q, [], 2);
            
        end
        
        clean_up;
        
        [loss, lossA, lossB, lossC] = compute_loss(x, z, K, U, L, mu, d, lambda, s);
        if (quiet ~= 1)
            fprintf('Iter %d, loss = %f, lossA = %f, lossB = %f, lossC = %f, K = %d\n', T, loss, lossA, lossB, lossC, K);
            disp(d(1:K)');
            disp(nk(1:K));
        end
        
    end
    
    w_ret = zeros(K, D, D);
    mu_ret = zeros(K, D);
    d_ret = zeros(K, 1);
    
    for k = 1:K
        mu_ret(k, :) = mu(k, :);
        d_ret(k) = d(k);
        if (d(k) > 0)
            dk = d(k);
            Uk = squeeze(U(k, :, :)); % D x D
            tL = zeros(dk, dk);
            for j = 1:dk
                tL(j, j) = L(k, j);
            end
            Wk = Uk(:, 1:dk) * tL; % D x dk
            w_ret(k, :, 1:dk) = Wk;
        end
    end

end