function [ z, Pi, W, mu, sigmasq ] = em_mppca( x, K, d, maxT, eps )
% Input:
%   x: real n x D, the data
%   K: int scalar, # of components
%   d: int K x 1, subspace dimensions
%
% Output:
%   z: int n x 1, hard clustering result (MAP estimate)
%   Pi: real K x 1, prior mixtures
%   W: real K x D x D, transformation matrix
%   mu: real K x D, offset
%   sigmasq: real K x 1, sigma^2 for each cluster

    % Scales
    n = size(x, 1);
    D = size(x, 2);
    
    % Initialization
    Pi = 1.0/K * ones(K, 1);
    W = zeros(K, D, D);
    for k = 1:K
        W(k, :, 1:d(k)) = randn(D, d(k));
    end
    mu = randn(K, D);
    sigmasq = ones(K, 1);
    
    lastloss = 1e100;
    loss = compute_loss_mppca(x, K, d, Pi, W, mu, sigmasq);
    fprintf('loss = %f\n', loss);
    
    % The EM algorithm
    
    T = 0;
    while (T < maxT && abs(lastloss - loss) / lastloss > eps)
        
        T = T + 1;
        lastloss = loss;
        fprintf('Iter %d: ', T);
        
        W0 = W;
        R = compute_posterior_mppca(x, K, d, Pi, W, mu, sigmasq); % n x K
        
        for k = 1:K
            
            % update pi
            Rk = R(:, k); % n x 1
            Pi(k) = 1.0/n * sum(Rk);
            
            % update mu
            muk = sum(Rk(:, ones(1, D)) .* x) / sum(Rk); % 1 x D
            mu(k, :) = muk;
            
            % compute the covariance matrix
            A = Rk(:, ones(1, D)) .* (x - muk(ones(n, 1), :)); % n x D
            B = x - muk(ones(n, 1), :); % n x D
            Sk = 1.0/(Pi(k)*n) * (B' * A); % D x D
            
            % compute Mk
            W0k = squeeze(W0(k, :, 1:d(k))); % D x d
            if (size(W0k, 1) ~= D)
                W0k = W0k';
            end            
            Mk = sigmasq(k) * eye(d(k)) + W0k' * W0k; % d x d
            
            % update W
            Wk = Sk * W0k * (sigmasq(k) * eye(d(k)) + Mk^(-1) * W0k' * Sk * W0k)^(-1); % D x d
            W(k, :, 1:d(k)) = Wk;
            
            % update sigma
            sigmasq(k) = 1.0/D * trace(Sk - Sk * W0k * Mk^(-1) * Wk');
            
        end
        
        loss = compute_loss_mppca(x, K, d, Pi, W, mu, sigmasq);
        fprintf('loss = %f\n', loss);
        
    end
    
    % make predictions
    R = compute_posterior_mppca(x, K, d, Pi, W, mu, sigmasq);
    [max_value, z] = max(R, [], 2);

end

